# -*- coding: utf-8 -*-
import tensorflow as tf
from tensorflow import keras


class KerasFeatureInputMergedModel(keras.Model):
    """
    keras多输入多输出设计: 合并几个模型一起训练
    注意：如果从tensorflow-2.0.0alpha0升级到tensorflow=2.0.0-beta1时会报错: tensorflow.python.framework.errors_impl.InvalidArgumentError:  You must feed a value for placeholder tensor 'child_gender_young'
    work around for issue: Unable to use FeatureColumn with Keras Functional API #27416  https://github.com/tensorflow/tensorflow/issues/27416
    https://www.curiousily.com/posts/heart-disease-prediction-in-tensorflow-2/
    ```
    import pandas as pd
    from tensorflow import feature_column
    from tensorflow import keras

    feature_columns = []
    child_month_young = feature_column.numeric_column("child_month_young")
    kid_age_youngest_buckets = feature_column.bucketized_column(child_month_young, boundaries=[0, 12, 24, 36, 72, 96])
    feature_columns.append(kid_age_youngest_buckets)

    child_gender_young = feature_column.categorical_column_with_vocabulary_list(
        'child_gender_young', ['boy', 'girl', ''])
    child_gender_young_one_hot = feature_column.indicator_column(child_gender_young)
    feature_columns.append(child_gender_young_one_hot)

    def keras_merged_model_fn(train_dataset, feature_columns, feature_column_names=["child_month_young", "child_gender_young"]):
        # first model
        first_input = keras.layers.Input(shape=(1,), name='first_input')
        first_input_encoded = keras.layers.Embedding(input_dim=10, output_dim=3, input_length=1)(first_input)
        first_input_encoded = keras.layers.Reshape((3,))(first_input_encoded)

        # second model
        second_input = keras.layers.Input(shape=(1,), name='second_input')
        second_input_encoded = keras.layers.Embedding(input_dim=10, output_dim=3, input_length=1)(second_input)
        second_input_encoded = keras.layers.Reshape((3,))(second_input_encoded)

        # merged model
        keras_feature_column_output_layer, keras_pure_input_layer_dict = \
            KerasFeatureInputMergedModel.build_keras_feature_column_output_layer(train_dataset.take(1),
                                                                            feature_columns,
                                                                            feature_column_names)
        merged_outputs = keras.layers.concatenate([first_input_encoded, second_input_encoded, keras_feature_column_output_layer])
        merged_outputs = keras.layers.Dense(1)(merged_outputs)

        merged_inputs = [first_input, second_input] + list(keras_pure_input_layer_dict.values())
        merged_model = keras.Model(inputs=merged_inputs, outputs=merged_outputs)
        return merged_model

    df = pd.DataFrame(
        [[3, 4, 5, 'boy', 0],
         [2, 6, 7, 'girl', 1]],
        columns=['first_input', 'second_input', 'child_month_young', 'child_gender_young', 'target'])

    def df_to_dataset(dataframe, shuffle=True, batch_size=32):
        dataframe = dataframe.copy()
        labels = dataframe.pop('target')
        d = dict(dataframe)
        ds = tf.data.Dataset.from_tensor_slices((d, labels))
        if shuffle:
            ds = ds.shuffle(buffer_size=len(dataframe))
        ds = ds.batch(batch_size)
        return ds

    train_ds = df_to_dataset(df)
    val_ds = df_to_dataset(df)

    merged_model = keras_merged_model_fn(train_ds, feature_columns, feature_column_names=["child_month_young", "child_gender_young"])
    merged_model.compile(optimizer='adam', loss='binary_crossentropy', metrics=['accuracy'])
    merged_model.fit(train_ds, validation_data=val_ds, epochs=1)
    #######################################
    # model.save("/tmp/saved_model/xxx.h5")
    # model.load_weights("/tmp/saved_model/xxx.h5")
    ```
    """
    def __init__(self, *args, **kwargs):
        keras_feature_columns = kwargs.pop('feature_columns', [])
        keras_feature_column_names = kwargs.pop('feature_column_names', [])
        super().__init__(*args, **kwargs)
        self.keras_feature_columns = keras_feature_columns
        self.keras_feature_column_names = keras_feature_column_names

    def call(self, inputs, training=False):
        from tensorflow.python.keras.engine import training_utils
        feature_input_items = {k: v for k, v in inputs.items() if k in self.keras_feature_column_names}
        self.feature_inputs = training_utils.ModelInputs(feature_input_items).get_symbolic_inputs()
        return tf.keras.layers.DenseFeatures(self.keras_feature_columns)(inputs)

    @staticmethod
    def build_keras_feature_column_output_layer(one_length_train_dataset, feature_columns, feature_column_names):
        feature_model = KerasFeatureInputMergedModel(feature_columns=feature_columns, feature_column_names=feature_column_names)
        feature_model.compile(optimizer='adam', loss='binary_crossentropy')
        feature_model.fit(one_length_train_dataset, epochs=1)
        fitted_feature_inputs = feature_model.feature_inputs

        keras_pure_input_layer_dict = {}
        for name, tensor in fitted_feature_inputs.items():
            keras_pure_input_layer = keras.layers.Input(shape=tensor.shape, name=name, tensor=tensor)
            keras_pure_input_layer_dict[name] = keras_pure_input_layer
        return tf.keras.layers.DenseFeatures(feature_columns)(keras_pure_input_layer_dict), keras_pure_input_layer_dict




# -*- coding: utf-8 -*-
import tensorflow as tf
from tensorflow_estimator import estimator  # estimator import方便查看源代码
from skydl.model.default_estimator_net import DefaultEstimatorNet
from skydl.model.default_model import DefaultModel


class DefaultEstimatorModel(DefaultModel):
    """
    default estimator model implements
    要从编译的Keras模型创建Estimator，我们需要调用model_to_estimator方法。
    请注意，Keras模型的初始模型状态将保留在创建好Estimator中。
    那么Estimator有什么好处呢？那么开始：
    1.无需更改模型即可在本地主机或分布式多GPU环境中运行基于Estimator的模型；
    2.Estimator简化了模型开发人员之间的共享实现；
    3.Estimator为你构建了图，有点像Eager Execution，没有显式会话。
    """
    @property
    def model(self)->estimator.Estimator:
        """定义estimator.Estimator返回值类型便于IDE找到对应的静态类"""
        return self._model

    def __init__(self, name=None, layers=None, model_fn=None, params=None, loss=None, optimizer=None, metrics=None, weights=None):
        """参数model_fn返回的对象可以是keras.Model的subclass实例. 也可以是estimator.EstimatorSpec实例"""
        self._layers = layers
        self._model_fn = model_fn
        self._params = params
        super().__init__(name=name, loss=loss, optimizer=optimizer, metrics=metrics, weights=weights)

    def build_network(self):
        return DefaultEstimatorNet(self.name if self.name else self.__class__.__name__, self.parser_args,
                                   self._layers, self._model_fn, self._params)

    def load_data(self):
        pass

    def compile(self, loss=None, optimizer=None, metrics=None):
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics
        super().compile(loss=self.loss, optimizer=self.optimizer, metrics=self.metrics)
        # restore weights from latest checkpoint
        latest_checkpoint = tf.train.latest_checkpoint(self.get_model_checkpoint_dir())
        if latest_checkpoint:
            self.model.load_weights(latest_checkpoint)
        return self

    def fit(self):
        return self

    def evaluate(self, evaluate_dataset=None):
        return self

    def serving(self):
        return self




# -*- coding: utf-8 -*-
import tensorflow as tf
import tensorflow_datasets as tfds
import six
import csv
import collections
from six.moves import urllib
import pandas as pd
import numpy as np
from skydl.datasets.super_tf_dataset_builder import DefaultDatasetBuilderConfig, SuperDatasetBuilder
import os,sys
sys.path.append(os.path.abspath('%s/..' % sys.path[0]))

_LABEL_DICT = collections.OrderedDict([
    ("1", "1st_class"), ("2", "2nd_class"), ("3", "3rd_class"), ("4", "4rd_class"), ("5", "5rd_class")
])


class MovieLensDatasetBuilder(SuperDatasetBuilder):
    """
    参考 tensorflow-datasets/mnist.py
    https://github.com/tensorflow/models/blob/master/official/datasets/movielens.py
    """
    BUILDER_CONFIGS = [
        DefaultDatasetBuilderConfig(
            name="plain_text",
            version="0.0.15",
            description="Plain text",
        ),
        DefaultDatasetBuilderConfig(
            name="bytes",
            version="0.0.1",
            description=("Uses byte-level text encoding with "
                         "`tfds.features.text.ByteTextEncoder`"),
            text_encoder_config=tfds.features.text.TextEncoderConfig(
                encoder=tfds.features.text.ByteTextEncoder()),
        ),
        DefaultDatasetBuilderConfig(
            name="subwords8k",
            version="0.0.1",
            description=("Uses `tfds.features.text.SubwordTextEncoder` with 8k "
                         "vocab size"),
            text_encoder_config=tfds.features.text.TextEncoderConfig(
                encoder_cls=tfds.features.text.SubwordTextEncoder,
                vocab_size=2 ** 13),
        ),
        DefaultDatasetBuilderConfig(
            name="subwords32k",
            version="0.0.1",
            description=("Uses `tfds.features.text.SubwordTextEncoder` with "
                         "32k vocab size"),
            text_encoder_config=tfds.features.text.TextEncoderConfig(
                encoder_cls=tfds.features.text.SubwordTextEncoder,
                vocab_size=2 ** 15),
        ),
    ]

    RATINGS_FILE = "ratings"
    MOVIES_FILE = "movies"
    USERS_FILE = "users"
    JOINT_FILE = "joint"
    GENRE_COLUMN = "genres"
    ITEM_COLUMN = "item_id"  # movies
    RATING_COLUMN = "rating"
    TIMESTAMP_COLUMN = "timestamp"
    TITLE_COLUMN = "titles"
    USER_COLUMN = "user_id"
    GENRES = [
        'Action', 'Adventure', 'Animation', "Children's", 'Comedy', 'Crime',
        'Documentary', 'Drama', 'Fantasy', 'Film-Noir', 'Horror', "IMAX", 'Musical',
        'Mystery', 'Romance', 'Sci-Fi', 'Thriller', 'War', 'Western'
    ]
    GENRES_CATEGORICAL_FEATURE_COLUMN = SuperDatasetBuilder.build_one_hot_feature_column(SuperDatasetBuilder.build_categorical_column_with_vocabulary_list(
                                                                                "genres", GENRES))
    N_GENRE = len(GENRES)

    RATING_COLUMNS = [USER_COLUMN, ITEM_COLUMN, RATING_COLUMN, TIMESTAMP_COLUMN]
    MOVIE_COLUMNS = [ITEM_COLUMN, TITLE_COLUMN, GENRE_COLUMN]
    USER_COLUMNS = [USER_COLUMN, 'gender', 'age', 'occupation', 'zip']

    RATING_FEATURE_DICT = collections.OrderedDict([
        ("user_id", (tf.int32, SuperDatasetBuilder.convert_to_int)),
        ("gender", (tf.string, SuperDatasetBuilder.convert_to_string)),
        ("age", (tf.int32, SuperDatasetBuilder.convert_to_int)),
        ("occupation", (tf.int32, SuperDatasetBuilder.convert_to_int)),
        ("zip", (tf.string, SuperDatasetBuilder.convert_to_string)),
        ("item_id", (tf.int32, SuperDatasetBuilder.convert_to_int)),
        ("timestamp", (tf.int32, SuperDatasetBuilder.convert_to_int)),
        # ("genres", (tf.string, SuperDatasetBuilder.convert_to_string))
        ("genres", (tfds.features.Tensor(dtype=tf.int64, shape=(None,)), SuperDatasetBuilder.convert_to_int_from_string))  # genres是不定长的一维数组，e.g. [7] or [14,23,55]
    ])

    def __init__(self, data_dir=None, config=None):
        super().__init__(data_dir=data_dir, config=config)
        self._num_shards = 1
        self._train_data_file = "http://files.grouplens.org/datasets/movielens/"
        self._test_data_file = None
        self._validation_data_file = None

    def _info(self):
        return tfds.core.DatasetInfo(
            builder=self,
            description=("movielens dataset"),
            features=tfds.features.FeaturesDict({
                "label": tfds.features.ClassLabel(names=[label for label in _LABEL_DICT.values()]),
                "feature": {name: dtype
                            for name, (dtype, func) in self.RATING_FEATURE_DICT.items()},
            }),
            supervised_keys=("feature", "label"),
            urls=["https://xxx"],
            citation="@ONLINE {movielens dataset}"
        )

    def _split_generators(self, dl_manager):
        filenames = {
            "ml-1m": "ml-1m.zip",
        }
        generator_array = []
        if self._train_data_file:
            # added checksums file: tensorflow_datasets/url_checksums/movie_lens_dataset_builder.txt
            dl_manager._register_checksums = True
            data_files = dl_manager.download_and_extract(
                {k: urllib.parse.urljoin(self._train_data_file, v) for k, v in filenames.items()})
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TRAIN,
                num_shards=3,
                gen_kwargs={
                    "data_path": data_files['ml-1m']
                }))
        if self._test_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TEST,
                num_shards=1,
                gen_kwargs={
                    "data_path": self._test_data_file
                }))
        if self._validation_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.VALIDATION,
                num_shards=1,
                gen_kwargs={
                    "data_path": self._validation_data_file
                }))
        return generator_array

    def _generate_examples(self, **kwargs):
        """Generate features and target given the directory path.
        Args:
          file_path: path where the csv file is stored
        Yields:
          The features and the target
        """
        data_path = kwargs.get("data_path")
        for dataset_path_name in tf.io.gfile.listdir(data_path):
            self._transform_csv(
                input_path=data_path + "/" + dataset_path_name + "/" + self.RATINGS_FILE + ".dat",
                output_path=data_path + "/" + dataset_path_name + "/" + self.RATINGS_FILE + ".csv",
                names=self.RATING_COLUMNS, skip_first=False, separator="::")
            self._transform_csv(
                input_path=data_path + "/" + dataset_path_name + "/" + self.MOVIES_FILE + ".dat",
                output_path=data_path + "/" + dataset_path_name + "/" + self.MOVIES_FILE + ".csv",
                names=self.MOVIE_COLUMNS, skip_first=False, separator="::")
            self._transform_csv(
                input_path=data_path + "/" + dataset_path_name + "/" + self.USERS_FILE + ".dat",
                output_path=data_path + "/" + dataset_path_name + "/" + self.USERS_FILE + ".csv",
                names=self.USER_COLUMNS, skip_first=False, separator="::")
            joint_pd = self.csv_to_joint_dataframe(data_path, dataset_path_name)
            joint_pd.to_csv(data_path + "/" + dataset_path_name + "/" + self.JOINT_FILE + ".csv", index=False)
        with tf.io.gfile.GFile(data_path + "/" + dataset_path_name + "/" + self.JOINT_FILE + ".csv") as f:
            raw_data = csv.DictReader(f)
            for row in raw_data:
                label_val = row.pop(self.RATING_COLUMN)
                yield {
                    "label": SuperDatasetBuilder.convert_to_label(str(int(float(label_val))), _LABEL_DICT),
                    "feature": {
                        name: self.RATING_FEATURE_DICT[name][1](value)
                        for name, value in row.items()
                    }
                }

    def _transform_csv(self, input_path, output_path, names, skip_first, separator=","):
      """Transform csv to a regularized format.
      Args:
        input_path: The path of the raw csv.
        output_path: The path of the cleaned csv.
        names: The csv column names.
        skip_first: Boolean of whether to skip the first line of the raw csv.
        separator: Character used to separate fields in the raw csv.
      """
      if six.PY2:
        names = [n.decode("utf-8") for n in names]
      with tf.io.gfile.GFile(output_path, "wb") as f_out, \
          tf.io.gfile.GFile(input_path, "rb") as f_in:
        # Write column names to the csv.
        f_out.write(",".join(names).encode("utf-8"))
        f_out.write(b"\n")
        for i, line in enumerate(f_in):
          if i == 0 and skip_first:
            continue  # ignore existing labels in the csv
          line = line.decode("utf-8", errors="ignore")
          fields = line.split(separator)
          if separator != ",":
            fields = ['"{}"'.format(field) if "," in field else field
                      for field in fields]
          f_out.write(",".join(fields).encode("utf-8"))

    def csv_to_joint_dataframe(self, data_dir, dataset):
        with tf.io.gfile.GFile(os.path.join(data_dir, dataset, self.RATINGS_FILE + ".csv")) as f_rating:
            ratings = pd.read_csv(f_rating, encoding="utf-8")
            with tf.io.gfile.GFile(os.path.join(data_dir, dataset, self.MOVIES_FILE + ".csv")) as f_item:
                movies = pd.read_csv(f_item, encoding="utf-8")
                df = ratings.merge(movies, on=self.ITEM_COLUMN)
                df[self.RATING_COLUMN] = df[self.RATING_COLUMN].astype(np.float32)
                df = df.drop(columns=[self.TITLE_COLUMN])
                df = self.integerize_genres(dataframe=df)
                ratings = df
            with tf.io.gfile.GFile(os.path.join(data_dir, dataset, self.USERS_FILE + ".csv")) as f_user:
                users = pd.read_csv(f_user, encoding="utf-8")
                df = ratings.merge(users, on=self.USER_COLUMN)
                df[self.RATING_COLUMN] = df[self.RATING_COLUMN].astype(np.float32)
                return df

    def integerize_genres(self, dataframe):
        """Replace genre string with a binary vector.
        Args:
          dataframe: a pandas dataframe of movie data.
        Returns:
          The transformed dataframe.
        """
        def _map_fn(entry):
            # entry.replace("Children's", "Children")  # naming difference.
            movie_genres = entry.split("|")
            output = np.zeros((len(self.GENRES),), dtype=np.int64)
            for i, genre in enumerate(self.GENRES):
                if genre in movie_genres:
                    output[i] += 1  # to one-hot vector
            return output

        def _map_fn2(entry):
            # entry.replace("Children's", "Children")  # naming difference.
            movie_genres = entry.split("|")
            num_total = 30
            output = []
            for i, genre in enumerate(self.GENRES):
                if genre in movie_genres:
                    output.append(str(i+1))   # padding 0, 有意义的下标从1开始。to [id1, id2] for embedding as: [[-0.19328298  0.1491687  -0.24025935],[0.22094809 -0.09602437 -0.01006811]]
            output = output + ['0' for i in range(len(output), num_total)]  # fill with '0' to num_total elements
            output = ",".join(output)  # e.g. convert ['12','33'] to '12,33'
            return output
        dataframe[self.GENRE_COLUMN] = dataframe[self.GENRE_COLUMN].apply(_map_fn2)
        return dataframe


if __name__ == '__main__':
    from tensorflow.python import feature_column
    embedding_featrues = {
        'price': [[1.0], [5.0], [10.0]],
        'color': [['R'], ['G'], ['B']]
    }
    price_column = feature_column.numeric_column('price')
    color_column = feature_column.categorical_column_with_vocabulary_list('color', ['R', 'G', 'B'])
    prediction = SuperDatasetBuilder.build_linear_model(embedding_featrues, [price_column, color_column]).numpy()
    print("feature_values" + "***"*40 + "\n", prediction)
    def demo_embedding():
        color_data = {"color": ["R", "B"]}
        color_column = SuperDatasetBuilder.build_categorical_column_with_vocabulary_list("color", ["R", "G", "B"])
        color_embeding = SuperDatasetBuilder.build_embedding_feature_column(color_column, 8)
        color_embeding_value = feature_column.input_layer(color_data, [color_embeding]).numpy()
        print("embedding" + "*" * 40)
        print(color_embeding_value)
    demo_embedding()

    print(SuperDatasetBuilder.list_builders())
    # build feature_columns
    feature_columns = []
    # feature_columns = MovieLensDatasetBuilder.build_numeric_feature_columns(MovieLensDatasetBuilder.camelcase_to_snakecase("MovieLensDatasetBuilder"))
    feature_columns.append(MovieLensDatasetBuilder.build_one_hot_feature_column(MovieLensDatasetBuilder.build_categorical_column_with_vocabulary_list("genres", MovieLensDatasetBuilder.GENRES)))
    # load datasets
    batched_datasets = MovieLensDatasetBuilder.load_batched_datasets(MovieLensDatasetBuilder.camelcase_to_snakecase("MovieLensDatasetBuilder"), config_name="plain_text")
    batched_dataset_features = MovieLensDatasetBuilder.read_batched_datasets(batched_datasets)
    for features, labels in batched_dataset_features:
        # print(features['genres'], tf.keras.layers.DenseFeatures(feature_columns)(features).numpy())
        print("features['genres']=", features['genres'])
    print("end!!")








# coding: utf-8 or # -*- coding: utf-8 -*-
from enum import Enum, unique


@unique
class TrainPhaseEnum(Enum):
    """
    train->test->free|serving->inference
    TODO 注意：后面版本应该废弃free|serving中间状态
    """
    Train = 'train'  # train phase
    Test = 'test'  # test phase(i.e. evaluation phase)
    Freeze = 'freeze'  # freeze model(test之后，将graph中变量参数值固化成常量)
    Serving = 'serving'  # serving model(test之后，将模型预测能力对外暴露成API服务模式)
    Inference = 'inference'  # online inference phase


if __name__ == '__main__':
    for name, phase in TrainPhaseEnum.__members__.items():
        print(name, '=>', phase, ',', phase.value)



# -*- coding: utf-8 -*-


class Constants(object):
    """
    系统常量
    """
    # dict_load_data
    DICT_LOAD_DATA_LAZY_LOAD_DATA = "lazy_load_data"
    DICT_LOAD_DATA_ITER_TRAIN_DATA_LABELS = "iter_train_data_labels"
    DICT_LOAD_DATA_TOTAL_TRAIN_BATCH = "total_train_batch"
    DICT_LOAD_DATA_TRAIN_DATA = "train_data"
    DICT_LOAD_DATA_TEST_DATA = "test_data"
    DICT_LOAD_DATA_VALIDATION_DATA = "validation_data"
    DICT_LOAD_DATA_NUM_TRAIN_DATA = "num_train_data"
    DICT_LOAD_DATA_NUM_TEST_DATA = "num_test_data"
    DICT_LOAD_DATA_NUM_VALIDATION_DATA = "num_validation_data"
    DICT_LOAD_DATA_SHAPE_CHANNEL_HEIGHT_WIDTH = "shape_channel_height_width"






# -*- coding: utf-8 -*-
import tensorflow as tf
import tensorflow_datasets as tfds
import numpy as np
from skydl.model.default_keras_net import DefaultKerasNet
from skydl.model.default_model import DefaultModel
from tensorflow.python import keras  # 或from tensorflow import keras
from tensorflow_estimator import estimator  # estimator import方便查看源代码


class DefaultKerasModel(DefaultModel):
    """
    default keras model implements
    find a issue: Unable to use FeatureColumn with Keras Functional API #27416  https://github.com/tensorflow/tensorflow/issues/27416
    """
    @property
    def model(self)->keras.Sequential:
        """定义keras.Model返回值类型便于IDE找到对应的静态类"""
        return self._model

    def __init__(self, name=None, keras_layers=[], keras_inputs=None, keras_outputs=None, loss=None, optimizer=None, metrics=None, weights=None):
        self._keras_layers = keras_layers
        self._keras_inputs = keras_inputs
        self._keras_outputs = keras_outputs
        super().__init__(name=name, loss=loss, optimizer=optimizer, metrics=metrics, weights=weights)

    def build_network(self):
        return DefaultKerasNet(self.name if self.name else self.__class__.__name__, self.parser_args,
                               self._keras_layers, keras_inputs=self._keras_inputs, keras_outputs=self._keras_outputs)

    def load_data(self, *args, **kwargs):
        # return batched_train_dataset, batched_validation_dataset, batched_evaluate_dataset
        pass

    def compile(self, loss=keras.losses.sparse_categorical_crossentropy, optimizer=keras.optimizers.Adam(), metrics=["accuracy"]):
        if not self.is_training_phase() and not self.is_evaluation_phase():
            return self
        # with self.distribute_strategy.scope():  报错 ValueError: We currently do not support distribution strategy with a `Sequential` model that is created without `input_shape`/`input_dim` set in its first layer or a subclassed model.
        # with self.distribute_strategy.scope():
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics
        super().compile(loss=self.loss, optimizer=self.optimizer, metrics=self.metrics)
        # restore weights from latest checkpoint
        latest_checkpoint = tf.train.latest_checkpoint(self.get_model_checkpoint_dir())
        if latest_checkpoint:
            self.model.get_proxy().load_weights(latest_checkpoint)
            self.log.info("already called load_weights from: " + latest_checkpoint)
        self.model.get_proxy().compile(optimizer=self.optimizer,
                               loss=self.loss,
                               metrics=self.metrics)
        return self

    def model_to_estimator(self,keras_model_path=None,
                           custom_objects=None,
                           model_dir=None,
                           config=None) -> estimator.Estimator:
        """
        从编译的Keras模型创建Estimator，需要调用model_to_estimator方法
        注意：
        需要先执行compile()
        keras model subclass必须继承于超类keras.Sequential才能成功执行model_to_estimator()
        :return: estimator
        """
        # config = tf.estimator.RunConfig(train_distribute=self.distribute_strategy) # 报错：ValueError: Only TensorFlow native optimizers are supported with DistributionStrategy.
        # estimator = tf.keras.estimator.model_to_estimator(self.model, config=config)
        return keras.estimator.model_to_estimator(keras_model=self.model,
                                                   keras_model_path=keras_model_path,
                                                   custom_objects=custom_objects,
                                                   model_dir=model_dir,
                                                   config=config)

    def fit_callbacks(self):
        """
        keras fit callbacks
        :return:
        """
        # Function for decaying the learning rate.
        # You can define any decay function you need.
        def decay(epoch):
            if epoch < 3:
                return 1e-3
            elif 3 <= epoch < 7:
                return 1e-4
            else:
                return 1e-5

        class PrintLR(tf.keras.callbacks.Callback):
            def on_epoch_end(self, epoch, logs=None):
                print('\nLearning rate for epoch {} is {}'.format(epoch+1, self.model.optimizer.lr.numpy()))  # this self is inner class self, so no need set self.model.get_proxy()
        callbacks = [
            tf.keras.callbacks.TensorBoard(log_dir='./logs', histogram_freq=0, write_graph=True, write_images=True),
            keras.callbacks.ModelCheckpoint(self.get_model_checkpoint_dir() + "/" + self.latest_model_filename,
                                            verbose=1,
                                            save_weights_only=True),
            keras.callbacks.LearningRateScheduler(decay),
            PrintLR()
        ]
        return callbacks

    def fit(self, *args, **kwargs):
        return self

    def evaluate(self, *args, **kwargs):
        return self

    def serving(self, *args, **kwargs):
        # restore weights from latest checkpoint
        latest_checkpoint = tf.train.latest_checkpoint(self.get_model_checkpoint_dir())
        if latest_checkpoint:
            self.model.get_proxy().load_weights(latest_checkpoint)
            self.log.info("already called load_weights from: " + latest_checkpoint)
        saved_serving_path = self.parser_args.model_checkpoint_path + "/serving/models/" + self.model.name() + "/" + self.parser_args.model_version

        _, evaluate_dataset, _ = self.load_data()
        # prediction mode
        for batched_examples in tfds.as_numpy(evaluate_dataset.take(1)):
            test_batched_features, test_batched_labels = batched_examples
        predictions = self.model.get_proxy().predict(test_batched_features)
        print("serving predict, real:", np.argmax(predictions[0]), test_batched_labels[0])

        # save model for tf serving
        if not self.model.get_proxy().inputs:
            for batched_examples in tfds.as_numpy(evaluate_dataset.take(1)):
                test_batched_features, test_batched_labels = batched_examples
                self.model.get_proxy()._set_inputs(test_batched_features)
        tf.saved_model.save(self.model.get_proxy(), saved_serving_path)
        return self

    def predict(self, *args, **kwargs):
        """python版的预测接口，返回预测结果"""
        return None




# -*- coding: utf-8 -*-
from typing import Dict
import tensorflow_datasets as tfds
import tensorflow as tf
from tensorflow import keras as k
from tensorflow.python.ops import control_flow_util
"""
ref: https://github.com/tensorflow/tensorflow/issues/25414
Implement DCGAN using the new TF 2.0 API.

Also test tensorflow-datasets.

Celeb-A dataset.
"""


control_flow_util.ENABLE_CONTROL_FLOW_V2 = True


@tf.function
def bce(x, label, label_smoothing=0.0):
    """Returns the discrete binary cross entropy between x and the discrete label
    Args:
        x: a 2D tensor
        label: the discrete label, aka, the distribution to match
        label_smoothing: if greater than zero, smooth the labels

    Returns:
        The binary cros entropy
    """
    return k.losses.BinaryCrossentropy()(tf.ones_like(x) * label, x)


def min_max(positive, negative, label_smoothing=0.0):
    """Returns the discriminator (min max) loss
    Args:
        positive: the discriminator output for the positive class: 2D tensor
        negative: the discriminator output for the negative class: 2D tensor
        smooth: if greater than zero, appiles one-sided label smoothing
    Returns:
        The sum of 2 BCE
    """
    one = tf.constant(1.0)
    zero = tf.constant(0.0)
    d_loss = bce(positive, one, label_smoothing) + bce(negative, zero)
    return d_loss


class Generator(k.Model):

    def __init__(self):
        super(Generator, self).__init__()
        self.fc1 = k.layers.Dense(4 * 4 * 1024)
        self.batchnorm1 = k.layers.BatchNormalization()

        self.conv2 = k.layers.Conv2DTranspose(
            filters=512,
            kernel_size=(5, 5),
            strides=(2, 2),
            padding="same",
            use_bias=False,
        )
        self.batchnorm2 = k.layers.BatchNormalization()

        self.conv3 = k.layers.Conv2DTranspose(
            filters=256,
            kernel_size=(5, 5),
            strides=(2, 2),
            padding="same",
            use_bias=False,
        )
        self.batchnorm3 = k.layers.BatchNormalization()

        self.conv4 = k.layers.Conv2DTranspose(
            filters=128,
            kernel_size=(5, 5),
            strides=(2, 2),
            padding="same",
            use_bias=False,
        )
        self.batchnorm4 = k.layers.BatchNormalization()

        self.conv5 = k.layers.Conv2DTranspose(
            filters=3,
            kernel_size=(5, 5),
            strides=(2, 2),
            padding="same",
            use_bias=False,
        )
        self.batchnorm5 = k.layers.BatchNormalization()

    def call(self, x: tf.Tensor, training: bool = True) -> tf.Tensor:
        x = self.fc1(x)
        x = self.batchnorm1(x, training=training)
        x = tf.nn.relu(x)
        x = tf.reshape(x, shape=(-1, 4, 4, 1024))

        x = self.conv2(x)
        x = self.batchnorm2(x, training=training)
        x = tf.nn.relu(x)

        x = self.conv3(x)
        x = self.batchnorm3(x, training=training)
        x = tf.nn.relu(x)

        x = self.conv4(x)
        x = self.batchnorm4(x, training=training)
        x = tf.nn.relu(x)

        x = self.conv5(x)
        x = self.batchnorm5(x, training=training)

        x = tf.nn.tanh(x)
        return x


class Discriminator(k.Model):
    def __init__(self):
        super(Discriminator, self).__init__()
        self.conv1 = k.layers.Conv2D(128, (5, 5), strides=(2, 2), padding="same")
        self.conv2 = k.layers.Conv2D(256, (5, 5), strides=(2, 2), padding="same")
        self.batchnorm2 = k.layers.BatchNormalization()
        self.conv3 = k.layers.Conv2D(512, (5, 5), strides=(2, 2), padding="same")
        self.batchnorm3 = k.layers.BatchNormalization()
        self.conv4 = k.layers.Conv2D(1024, (5, 5), strides=(2, 2), padding="same")
        self.batchnorm4 = k.layers.BatchNormalization()
        self.flatten = k.layers.Flatten()
        self.fc5 = k.layers.Dense(1)

    def call(self, x, training=True):
        x = self.conv1(x)
        x = tf.nn.leaky_relu(x)

        x = self.conv2(x)
        x = self.batchnorm2(x)
        x = tf.nn.leaky_relu(x)

        x = self.conv3(x)
        x = self.batchnorm3(x)
        x = tf.nn.leaky_relu(x)

        x = self.conv4(x)
        x = self.batchnorm4(x)
        x = tf.nn.leaky_relu(x)

        x = self.flatten(x)
        x = self.fc5(x)
        return x


class GAN:
    def __init__(self, generator, discriminator, encoder=None):
        """
        GAN initializer.

        Args:
            generator: A ``tensorflow.keras.Model`` to use as Generator.
            discriminator: A ``tensorflow.keras.Model`` to use as Discriminator.
            encoder: A ``tensorflow.keras.Model`` to use as Encoder.

        Returns:
            Trained GAN model (?).

        """
        self.G = generator()
        self.D = discriminator()
        # self.E = encoder() if encoder is not None else None
        self.latent_vector_dims = 100

    @tf.function()
    def train(self, dataset, opt1, opt2):
        """
        Train.
        """
        step = 0
        for f in dataset:
            x = f["image"]
            step += 1
            z = tf.random.normal((32, self.latent_vector_dims))

            # We record all the operations in the tape
            with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
                G_z = self.G(z, training=True)

                D_x = self.D(x, training=True)
                D_Gz = self.D(G_z, training=True)

                g_loss = bce(D_Gz, 1.0)
                d_loss = min_max(D_x, D_Gz, label_smoothing=0.0)

            # We retrieve the gradients from our records
            G_grads = gen_tape.gradient(g_loss, self.G.trainable_variables)
            D_grads = disc_tape.gradient(d_loss, self.D.trainable_variables)

            # Optimize and apply the gradients
            opt1.apply_gradients(zip(G_grads, self.G.trainable_variables))
            opt2.apply_gradients(zip(D_grads, self.D.trainable_variables))

            if step % 1 == 0:
                print("--------------------------")
                print("STEP", step)
                print("D_LOSS", d_loss)
                print("G_LOSS:", g_loss)
        return step


class InputPipeline:

    def __init__(self, dataset, batch_size, epochs, shuffle_buffer, prefetched_items, size):
        self.batch_size = batch_size
        self.dataset_name = dataset
        self.epochs = epochs
        self.prefetched_items = prefetched_items
        self.shuffle_buffer = shuffle_buffer
        self.size = size

    def get_input_fn(self):
        """Input fn."""
        return self.input_fn

    def load_public_dataset(self):
        """
        Load one of the publicly available datasets, will merge together all the splits.

        Args:
            chosen_dataset: dataset to use.

        Return:
            The chosen dataset as a ``tf.data.Dataset``

        """
        # Construct a tf.data.Dataset
        datasets = tfds.load(name=self.dataset_name, split=tfds.Split.ALL)
        return datasets

    def resize_images(self, features):
        """
        Overwrite the \"image\" feature in order to resize them.

        Args:
            features: features dictionary.
            size: desired target size.

        Returns:
            Features with \"image\" resized to the correct shape.

        """
        features["image"] = tf.image.resize(features["image"], self.size)
        return features

    def input_fn(self):
        dataset = self.load_public_dataset()
        dataset = (
            dataset.map(self.resize_images)
            .shuffle(self.shuffle_buffer)
            .batch(self.batch_size)
            .prefetch(self.prefetched_items)
            .repeat(self.epochs)
        )
        return dataset


def main():
    # See available datasets
    public_datasets = tfds.list_builders()
    print(public_datasets)

    gan = GAN(Generator, Discriminator)
    G_opt = k.optimizers.Adam(learning_rate=1e-5, beta_1=0.5)
    D_opt = k.optimizers.Adam(learning_rate=1e-5, beta_1=0.5)

    input_pipeline = InputPipeline(
        dataset="celeb_a",
        batch_size=32,
        epochs=2,
        prefetched_items=1,
        shuffle_buffer=1000,
        size=(64, 64),
    )
    dataset = input_pipeline.input_fn()
    _ = gan.train(dataset, G_opt, D_opt)


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
from skydl.model.default_estimator_model import DefaultEstimatorModel
from tfv2 import iris_data
#########################
# 方便查看源代码
from tensorflow_estimator import estimator
from tensorflow.python import feature_column


class MyEstimatorModel(DefaultEstimatorModel):
    """
    tensorflow estimator model
    """
    def adjust_parse_args_value(self):
        super().adjust_parse_args_value()

    def load_data(self):
        # (train_x, train_y), (test_x, test_y) = iris_data.load_data()
        return iris_data.load_data()

    def fit(self):
        from tfv2 import iris_data
        batch_size = 128
        train_steps = 10
        (train_x, train_y), (test_x, test_y) = self.load_data()
        # Feature columns describe how to use the input.
        my_feature_columns = []
        for key in train_x.keys():
            my_feature_columns.append(feature_column.numeric_column(key=key))
        self.model.train(
            input_fn=lambda: iris_data.train_input_fn(train_x, train_y, batch_size),
            steps=train_steps)
        eval_result = self.model.evaluate(input_fn=lambda: iris_data.eval_input_fn(test_x, test_y, batch_size))
        print('\nTest set accuracy: {accuracy:0.3f}\n'.format(**eval_result))
        return self

    def fit2(self):
        import numpy as np
        import pandas as pd
        from matplotlib import pyplot as plt
        import tensorflow as tf
        #########################
        # 方便查看源代码
        from tensorflow_estimator import estimator
        from tensorflow.python import feature_column
        #########################

        # Load dataset.
        # dftrain = pd.read_csv('~/tensorflow_datasets/manual_downloads/titanic/train.csv')
        # dfeval = pd.read_csv('~/tensorflow_datasets/manual_downloads/titanic/eval.csv')
        dftrain = pd.read_csv('https://storage.googleapis.com/tf-datasets/titanic/train.csv')
        dfeval = pd.read_csv('https://storage.googleapis.com/tf-datasets/titanic/eval.csv')

        y_train = dftrain.pop('survived')
        y_eval = dfeval.pop('survived')

        tf.random.set_seed(123)

        fc = feature_column
        CATEGORICAL_COLUMNS = ['sex', 'n_siblings_spouses', 'parch', 'class', 'deck',
                               'embark_town', 'alone']
        NUMERIC_COLUMNS = ['age', 'fare']

        def one_hot_cat_column(feature_name, vocab):
            return feature_column.indicator_column(
                feature_column.categorical_column_with_vocabulary_list(feature_name,
                                                                       vocab))

        feature_columns = []
        for feature_name in CATEGORICAL_COLUMNS:
            # Need to one-hot encode categorical features.
            vocabulary = dftrain[feature_name].unique()
            feature_columns.append(one_hot_cat_column(feature_name, vocabulary))

        for feature_name in NUMERIC_COLUMNS:
            feature_columns.append(feature_column.numeric_column(feature_name,
                                                                 dtype=tf.float32))

        example = dict(dftrain.head(1))
        class_fc = feature_column.indicator_column(
            feature_column.categorical_column_with_vocabulary_list('class', ('First', 'Second', 'Third')))
        print('Feature value: "{}"'.format(example['class'].iloc[0]))
        print('One-hot encoded: ', tf.keras.layers.DenseFeatures([class_fc])(example).numpy())

        print(tf.keras.layers.DenseFeatures(feature_columns)(example).numpy())

        # Use entire batch since this is such a small dataset.
        NUM_EXAMPLES = len(y_train)

        def make_input_fn(X, y, n_epochs=None, shuffle=True):
            def input_fn():
                dataset = tf.data.Dataset.from_tensor_slices((dict(X), y))
                if shuffle:
                    dataset = dataset.shuffle(NUM_EXAMPLES)
                # For training, cycle thru dataset as many times as need (n_epochs=None).
                dataset = dataset.repeat(n_epochs)
                # In memory training doesn't use batching.
                dataset = dataset.batch(NUM_EXAMPLES)
                return dataset
            return input_fn

        # Training and evaluation input functions.
        train_input_fn = make_input_fn(dftrain, y_train)
        eval_input_fn = make_input_fn(dfeval, y_eval, shuffle=False, n_epochs=1)
        self.model.train(train_input_fn, max_steps=100)
        print("eval...........")
        result = self.model.evaluate(eval_input_fn)
        print(pd.Series(result))
        return self

    def evaluate(self, *args, **kwargs):
        return self

    def serving(self):
        return self




# -*- coding: utf-8 -*-
import torch
import argparse
from skydl.model.super_model import SuperModel
from py_common_util.pytorch.torch_utils import TorchUtils
import numpy as np
from skydl.model.super_torch_net import SuperTorchNet
from logbook import Logger, StreamHandler, FileHandler
import os, sys
sys.path.append(os.path.abspath('%s/..' % sys.path[0]))


class SuperTorchModel(SuperModel):
    """
    pytorch model
    在线教程：https://github.com/zergtant/pytorch-handbook
    pytorch中文教程网：https://www.pytorchtutorial.com/
    Pytorch Seq2Seq篇 https://fgc.stpi.narl.org.tw/activity/videoDetail/4b1141305df38a7c015e194f22f8015b
    """
    @property
    def model(self)->torch.nn.Module:
        """定义keras.Model返回值类型便于IDE找到对应的静态类"""
        return self._model

    def __init__(self, name=None, loss=None, optimizer=None, metrics=None, weights=None):
        StreamHandler(sys.stdout).push_application()
        self._name = name
        self._log = Logger(self.name if self.name else self.__class__.__name__)
        self._parser = argparse.ArgumentParser()
        self._parser_args = None
        self._device = None
        self._num_gpu = 0
        self.do_parse_args()
        self._model = None if self.parser_args.lazy_load_model else self.build_network()
        # loss, optimizer, metrics, weights
        self._loss = loss
        self._optimizer = optimizer
        self._metrics = metrics
        self._weights = weights
        self._distribute_strategy = None

    def add_parser_argument(self):
        """
        增加parser参数项
        子类需要重写该方法
        # super().add_parser_argument()
        """
        self.parser.add_argument('--data_path', type=str, default=sys.path[0] + '/dataset/wikitext_2',
                            help='location of the data corpus')
        self.parser.add_argument('--onnx_export_path', type=str, default=sys.path[0] + '/saved_model',
                                 help='path to export the final model in onnx format')
        self.parser.add_argument('--log_path', type=str, default=sys.path[0] + '/logs',
                            help='location of the loading data corpus')
        self.parser.add_argument('--batch_size', type=int, default=64, metavar='N',
                            help='input batch size for training (default: 64)')
        self.parser.add_argument('--eval_batch_size', type=int, default=1000, metavar='N',
                            help='input batch size for testing (default: 1000)')
        self.parser.add_argument('--epochs', type=int, default=1, metavar='N',
                            help='number of epochs to train (default: 10)')
        self.parser.add_argument('--learning_rate', type=float, default=0.01, metavar='LR',
                            help='learning rate (default: 0.01)')
        self.parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                            help='SGD momentum (default: 0.5)')
        self.parser.add_argument('--use_cuda', action='store_true', default=True,
                            help='use CUDA')
        self.parser.add_argument('--seed', type=int, default=1, metavar='S',
                            help='random seed (default: 1)')
        self.parser.add_argument('--log_interval', type=int, default=50, metavar='N',
                            help='how many batches to wait before logging training status')
        self.parser.add_argument('--lazy_load_model', type=bool, default=False, metavar='N',
                             help='lazy load model(build_network())')

    def do_parse_args(self):
        self.add_parser_argument()
        self._parser_args = self.parser.parse_args()
        self.adjust_parse_args_value()
        # 设置random seed确保每次训练都可以获得相同唯一的的随机序列
        np.random.seed(self.parser_args.seed)
        torch.manual_seed(self.parser_args.seed)
        self._num_gpu = torch.cuda.device_count()
        if self.num_gpu > 0:
            if not self.parser_args.use_cuda:
                self.log.info("WARNING: You have " + str(self._num_gpu) + " CUDA device, so you should probably run with --cuda")
                # 不使用GPU加速
                self._num_gpu = 0
        else:
            self.parser_args.use_cuda = False
        self._device = TorchUtils.device("gpu" if self.parser_args.use_cuda else "cpu")
        self.log.info("PyTorch Version: " + torch.__version__ + "\nMain function argparse: " + str(self.parser_args))

    def build_network(self):
        return SuperTorchNet(self.name if self.name else self.__class__.__name__, self.parser_args).to(self.device)

    def load_data(self):
        pass

    def fit(self):
        pass

    def evaluate(self, evaluate_dataset=None):
        pass

    def serving(self):
        pass

    def predict(self, *args, **kwargs):
        pass








# -*- coding: utf-8 -*-
from skydl.model.torch_model_mixin import TorchModelMixin
from skydl.model.super_net import SuperNet


class SuperTorchNet(SuperNet, TorchModelMixin):
    """pytorch net"""
    def __init__(self, name=None, parser_args=None):
        super().__init__(name=name, parser_args=parser_args)
        # 再显式调用其它的__init__(self)
        TorchModelMixin.__init__(self)

    def forward(self, *input, **kwargs):
        return super().forward(*input, **kwargs)

    def call_hidden_layers(self, _x, name="predict"):
        name = 'predict' if name is None else name
        predict = None
        return predict

    def compile_loss_metrics_optimizer_predict(self, _x, _y, _learning_rate=0.001, devices=[]):
        pass





# -*- coding: utf-8 -*-
import tensorflow as tf
import tensorflow_datasets as tfds
import numpy as np
import sys, os
from skydl.datasets.movielens_dataset_builder import MovieLensDatasetBuilder
from skydl.model.default_model import DefaultModel
from tensorflow import keras
from py_common_util.common.common_utils import CommonUtils
from skydl.model.impl.movielens_wide_deep_keras_model import MovieLensWideDeepKerasModel
from skydl.model.impl.movielens_wide_deep_keras_model1 import MovieLensWideDeepKerasModel1
from skydl.model.impl.movielens_wide_deep_keras_model2 import MovieLensWideDeepKerasModel2
from skydl.model.keras_feature_input_layer_merged_model import KerasFeatureInputMergedModel
sys.path.append(os.path.abspath('%s/..' % sys.path[0]))


@CommonUtils.print_exec_time
def run_default_model():
    DefaultModel().fit()


@CommonUtils.print_exec_time
def run_keras_model():
    from skydl.model.impl.my_keras_model import MyKerasModel
    # feature_columns = [tf.feature_column.numeric_column("image", shape=[28, 28])]
    MyKerasModel("my_keras_model", [
        keras.layers.Flatten(input_shape=(28, 28)),
        keras.layers.Dropout(0.2),
        keras.layers.Dense(128, activation='relu'),
        keras.layers.Dense(10, activation='softmax', name="prediction")
    ]).compile(
        keras.losses.sparse_categorical_crossentropy,
        keras.optimizers.Adam(),
        ['accuracy']
        # ['accuracy', keras.metrics.Precision(), keras.metrics.Recall()]
    ).fit().serving()

    inference_model = MyKerasModel()
    _, evaluate_dataset, _ = inference_model.load_data()
    import tensorflow_datasets as tfds
    for dataset in tfds.as_numpy(evaluate_dataset):
        features, labels = dataset
    print("inference model name: ", inference_model.model.name())
    # inference_model.inference(model_name="my_keras_model",
    #                           features_list=features.tolist()[:2])
    # print("the inference results:", labels.tolist()[:2])


@CommonUtils.print_exec_time
def run_wine_wide_deep_keras_model():
    from skydl.model.impl.wine_wide_deep_keras_model import WineWideDeepKerasModel
    WineWideDeepKerasModel("wine_wide_deep_keras_model", [
        keras.layers.Flatten(input_shape=(28, 28)),
        keras.layers.Dropout(0.2),
        keras.layers.Dense(128, activation='relu'),
        keras.layers.Dense(10, activation='softmax', name="prediction")
    ]).compile(
        keras.losses.sparse_categorical_crossentropy,
        keras.optimizers.Adam(),
        ['accuracy']
    ).fit()


@CommonUtils.print_exec_time
def run_movielens_wide_deep_keras_model_ok():
    from skydl.model.impl.movielens_wide_deep_keras_model import MovieLensWideDeepKerasModel
    MovieLensWideDeepKerasModel("movielens_wide_deep_keras_model1", [
        keras.layers.Dense(units=128, activation='relu'),
        keras.layers.Flatten()
    ]).compile(
        loss="mean_squared_error",
        optimizer="adam",
        metrics=['accuracy']
    ).fit_ok()


@CommonUtils.print_exec_time
def run_movielens_wide_deep_keras_model():
    """
    1.Best Practices on Recommendation Systems https://github.com/microsoft/recommenders
    1.图书推荐，参考：Book Recommendation System https://github.com/TannerGilbert/Tutorials/blob/master/Recommendation%20System/Recommendation%20System.ipynb
    2.Building a book Recommendation System using Keras https://towardsdatascience.com/building-a-book-recommendation-system-using-keras-1fba34180699
    3.How to Implement a Recommendation System with Deep Learning and PyTorch https://medium.com/@iliazaitsev/how-to-implement-a-recommendation-system-with-deep-learning-and-pytorch-2d40476590f9
    """
    from skydl.model.impl.movielens_wide_deep_keras_model import MovieLensWideDeepKerasModel
    from skydl.datasets.movielens_dataset_builder import MovieLensDatasetBuilder
    dataset_name = MovieLensDatasetBuilder.camelcase_to_snakecase("MovieLensDatasetBuilder")
    num_examples = MovieLensDatasetBuilder.get_num_examples(dataset_name)
    full_datasets = MovieLensDatasetBuilder.load_batched_datasets(dataset_name,
                                                                  split=[tfds.Split.TRAIN],
                                                                  batch_size=128,
                                                                  epochs=1)
    train_dataset, validation_dataset, test_dataset = MovieLensDatasetBuilder.split_full_datasets(full_datasets,
                                                                                                  num_examples=MovieLensDatasetBuilder.get_total_num_batch(
                                                                                                      num_examples,
                                                                                                      128),
                                                                                                  train_size=0.7,
                                                                                                  validation_size=0,
                                                                                                  test_size=0.2)
    np_user_ids = np.array([])
    np_item_ids = np.array([])
    np_labels = np.array([])

    # train_dataset = full_datasets.take(100) # for dev
    for features, labels in MovieLensDatasetBuilder.read_batched_datasets(train_dataset.take(100)):
        np_user_ids = np.append(np_user_ids, features['user_id'])
        np_item_ids = np.append(np_item_ids, features['item_id'])
        np_labels = np.append(np_labels, labels)

    user_feature_column = MovieLensDatasetBuilder.build_numeric_feature_columns(["user_id"])
    item_feature_column = MovieLensDatasetBuilder.build_numeric_feature_columns(["item_id"])
    total_users = len(np.unique(np_user_ids))
    total_items = len(np.unique(np_item_ids))

    user_model = MovieLensWideDeepKerasModel("user_movielens_wide_deep_keras_model", [
        keras.layers.Input(shape=(1,), name="age"),
        keras.layers.Embedding(total_users, 10),
        keras.layers.Dense(units=128, activation='relu'),
        keras.layers.Flatten()
    ]).compile(
        loss="mean_squared_error",
        optimizer="adam",
        metrics=['accuracy']
    )

    item_model = MovieLensWideDeepKerasModel("item_movielens_wide_deep_keras_model", [
        keras.layers.Input(shape=(1,), name="timestamp"),
        keras.layers.Embedding(total_items, 10),
        keras.layers.Dense(units=128, activation='relu'),
        keras.layers.Flatten()
    ]).compile(
        loss="mean_squared_error",
        optimizer="adam",
        metrics=['accuracy']
    )

    # merged model
    keras_feature_column_input_layer, keras_pure_input_layer_dict = \
        KerasFeatureInputMergedModel.build_keras_feature_column_output_layer(train_dataset.take(1),
                                                                            user_feature_column+item_feature_column,
                                                                            ["user_id", "item_id"])
    merged_outputs = keras.layers.concatenate([user_model.model.get_static_logits_output(), item_model.model.get_static_logits_output(), keras_feature_column_input_layer])
    merged_outputs = keras.layers.Dense(1)(merged_outputs)
    merged_inputs = [user_model.model.get_static_first_input(), item_model.model.get_static_first_input()] + list(keras_pure_input_layer_dict.values())

    merged_model = MovieLensWideDeepKerasModel(
        "merged_movielens_wide_deep_keras_model",
        keras_inputs=merged_inputs,
        keras_outputs=merged_outputs
    ).compile(
        loss="binary_crossentropy",
        optimizer="adam",
        metrics=['accuracy']
    )
    merged_model.model.model_proxy.fit(train_dataset)
    merged_model.model.model_proxy.summary()


@CommonUtils.print_exec_time
def run_my_estimator_model():
    from tensorflow_estimator import estimator
    from tensorflow.python import feature_column
    from skydl.model.impl.my_estimator_model import MyEstimatorModel
    from tfv2 import iris_data
    batch_size = 128
    train_steps = 10

    # Fetch the data
    (train_x, train_y), (test_x, test_y) = iris_data.load_data()

    # Feature columns describe how to use the input.
    my_feature_columns = []
    for key in train_x.keys():
        my_feature_columns.append(feature_column.numeric_column(key=key))
    def my_model(features, labels, mode, params):
        """DNN with three hidden layers and learning_rate=0.1."""
        # Create three fully connected layers.
        net = feature_column.input_layer(features, params['feature_columns'])
        for units in params['hidden_units']:
            net = tf.compat.v1.layers.dense(net, units=units, activation=tf.nn.relu)
        # Compute logits (1 per class).
        logits = tf.compat.v1.layers.dense(net, params['n_classes'], activation=None)
        # Compute predictions.
        predicted_classes = tf.argmax(logits, 1)
        if mode == estimator.ModeKeys.PREDICT:
            predictions = {
                'class_ids': predicted_classes[:, tf.newaxis],
                'probabilities': tf.nn.softmax(logits),
                'logits': logits,
            }
            return estimator.EstimatorSpec(mode, predictions=predictions)
        # Compute loss.
        loss = tf.compat.v1.losses.sparse_softmax_cross_entropy(labels=labels, logits=logits)
        # Compute evaluation metrics.
        accuracy = tf.compat.v1.metrics.accuracy(labels=labels,
                                                 predictions=predicted_classes,
                                                 name='acc_op')
        metrics = {'accuracy': accuracy}
        tf.summary.scalar('accuracy', accuracy[1])
        if mode == estimator.ModeKeys.EVAL:
            return estimator.EstimatorSpec(mode, loss=loss, eval_metric_ops=metrics)
        # Create training op.
        assert mode == estimator.ModeKeys.TRAIN
        optimizer = tf.compat.v1.train.AdagradOptimizer(learning_rate=0.1)
        train_op = optimizer.minimize(loss, global_step=tf.compat.v1.train.get_global_step())
        return estimator.EstimatorSpec(mode, loss=loss, train_op=train_op)

    MyEstimatorModel("my_estimator_model",
                     model_fn=my_model,
                     params={
                        'feature_columns': my_feature_columns,
                        'hidden_units': [10, 10],
                        'n_classes': 3
                    }).compile().fit()


@CommonUtils.print_exec_time
def run_my_estimator_model2():
    import pandas as pd
    from tensorflow_estimator import estimator
    from tensorflow.python import feature_column
    from skydl.model.impl.my_estimator_model import MyEstimatorModel
    from skydl.datasets.super_tf_dataset_builder import SuperDatasetBuilder
    batch_size = 128
    train_steps = 10
    dftrain = pd.read_csv('https://storage.googleapis.com/tf-datasets/titanic/train.csv')
    dfeval = pd.read_csv('https://storage.googleapis.com/tf-datasets/titanic/eval.csv')

    y_train = dftrain.pop('survived')
    y_eval = dfeval.pop('survived')
    tf.random.set_seed(123)
    fc = feature_column
    CATEGORICAL_COLUMNS = ['sex', 'n_siblings_spouses', 'parch', 'class', 'deck', 'embark_town', 'alone']
    NUMERIC_COLUMNS = ['age', 'fare']

    feature_columns = []
    for feature_name in CATEGORICAL_COLUMNS:
        # Need to one-hot encode categorical features.
        vocabulary = dftrain[feature_name].unique()
        feature_columns.append(SuperDatasetBuilder.build_one_hot_column(SuperDatasetBuilder.build_categorical_column_with_vocabulary_list(feature_name, vocabulary)))

    for feature_name in NUMERIC_COLUMNS:
        feature_columns.append(SuperDatasetBuilder.build_numeric_column(feature_name, dtype=tf.float32))

    example = dict(dftrain.head(1))
    class_fc = SuperDatasetBuilder.build_one_hot_column(SuperDatasetBuilder.build_categorical_column_with_vocabulary_list('class', ('First', 'Second', 'Third')))
    print('Feature value: "{}"'.format(example['class'].iloc[0]))
    print('One-hot encoded: ', tf.keras.layers.DenseFeatures([class_fc])(example).numpy())

    est = estimator.BoostedTreesClassifier(feature_columns, n_batches_per_layer=batch_size)
    MyEstimatorModel("my_estimator_model2",
                     model_fn=est.model_fn).compile().fit2()


@CommonUtils.print_exec_time
def run_keras_model_to_estimator():
    from skydl.model.impl.my_keras_model import MyKerasModel
    model = MyKerasModel("my_keras_model_to_estimator", [
        keras.layers.Flatten(input_shape=(28, 28)),
        keras.layers.Dropout(0.2),
        keras.layers.Dense(128, activation='relu'),
        keras.layers.Dense(10, activation='softmax', name="prediction")
    ]).compile(
        keras.losses.sparse_categorical_crossentropy,
        keras.optimizers.Adam(),
        ['accuracy']
        # ['accuracy', keras.metrics.Precision(), keras.metrics.Recall()]
    )
    estimator = model.model_to_estimator()
    estimator.train(input_fn=lambda: model.load_data()[0])
    print("eval...........")
    result = estimator.evaluate(input_fn=lambda: model.load_data()[1])
    import pandas as pd
    print(pd.Series(result))


@CommonUtils.print_exec_time
def run_hst_recommend_system():
    """推荐系统：协同过滤"""
    from skydl.model.impl.movielens_wide_deep_keras_model import MovieLensWideDeepKerasModel
    train_dataset, validation_dataset, evaluate_dataset, np_user_ids, np_item_ids, np_labels = MovieLensWideDeepKerasModel().load_data()
    total_users = max(np.unique(np_user_ids))
    total_items = max(np.unique(np_item_ids))
    print("***total_users，total_items=***", total_users, total_items)  # 6040 3649

    user_model = MovieLensWideDeepKerasModel1("user_recommend", [
        keras.layers.Input(shape=(1,), name="user_id"),
        keras.layers.Embedding(6040, 10),
        keras.layers.Flatten()
    ])

    item_model = MovieLensWideDeepKerasModel2("item_recommend", [
        keras.layers.Input(shape=(1,), name="item_id"),
        keras.layers.Embedding(4000, 10),
        keras.layers.Flatten()
    ])

    # merged model
    merged_outputs = keras.layers.Dot(axes=1)([item_model.model.get_static_logits_output(), user_model.model.get_static_logits_output()])
    merged_outputs = keras.layers.Dense(1)(merged_outputs)
    merged_inputs = [user_model.model.get_static_first_input(), item_model.model.get_static_first_input()]

    print("inference instances: ", [{"user_id": np_user_ids[3], "item_id": np_item_ids[i]} for i in range(len(np_item_ids[:400]))])
    MovieLensWideDeepKerasModel(
        "user_book_recommend",
        keras_inputs=merged_inputs,
        keras_outputs=merged_outputs
    ).compile(
        loss="mean_squared_error",
        optimizer="adam",
        metrics=['accuracy']
    ).fit(
        loaded_dataset=[train_dataset, validation_dataset, evaluate_dataset, np_user_ids, np_item_ids, np_labels]
    ).serving(
    ).inference(tfx_host_port="http://192.168.83.111:8501/",
                model_name="user_book_recommend",
                features_list=[{"user_id": np_user_ids[3], "item_id": np_item_ids[i]} for i in range(len(np_item_ids[:400]))])


@CommonUtils.print_exec_time
def run_hst_recommend_system1():
    """推荐系统：协同过滤"""
    from skydl.model.impl.movielens_wide_deep_keras_model import MovieLensWideDeepKerasModel
    train_dataset, validation_dataset, evaluate_dataset, np_user_ids, np_item_ids, np_labels = MovieLensWideDeepKerasModel().load_data()
    total_users = max(np.unique(np_user_ids))
    total_items = max(np.unique(np_item_ids))
    print("***total_users，total_items=***", total_users, total_items)  # 6040 3649

    keras_feature_column_output_layer1, keras_pure_input_layer_dict1 = \
        KerasFeatureInputMergedModel.build_keras_feature_column_output_layer(train_dataset.take(1),
                                                                            MovieLensDatasetBuilder.build_numeric_feature_columns(["user_id"], dtype=tf.int32),
                                                                            ["user_id"])

    keras_feature_column_output_layer2, keras_pure_input_layer_dict2 = \
        KerasFeatureInputMergedModel.build_keras_feature_column_output_layer(train_dataset.take(1),
                                                                            MovieLensDatasetBuilder.build_numeric_feature_columns(["item_id"], dtype=tf.int32),
                                                                            ["item_id"])

    # keras_feature_column_output_layer3, keras_pure_input_layer_dict3 = \
    #     KerasFeatureInputMergedModel.build_keras_feature_column_output_layer(train_dataset.take(1),
    #                                                                         MovieLensDatasetBuilder.build_numeric_feature_columns(["genres"], dtype=tf.int64),
    #                                                                         ["genres"])
    keras_feature_column_output_layer3, keras_pure_input_layer_dict3 = \
        KerasFeatureInputMergedModel.build_keras_feature_column_output_layer(train_dataset.take(1),
                                                                            MovieLensDatasetBuilder.build_embedding_feature_column(MovieLensDatasetBuilder.build_categorical_column_with_identity("genres", num_buckets=30), 10),
                                                                            ["genres"])

    out1 = keras.layers.Flatten()(keras.layers.Embedding(6040, 10)(keras_feature_column_output_layer1))
    out2 = keras.layers.Flatten()(keras.layers.Embedding(4000, 10)(keras_feature_column_output_layer2))
    # out3 = keras.layers.Flatten()(keras.layers.Embedding(29+1, 10, mask_zero=False)(keras_feature_column_output_layer3))
    out3 = keras_feature_column_output_layer3

    # merged model
    merged_outputs = keras.layers.Dot(axes=1)([out1, out3])
    merged_outputs = keras.layers.Dense(1)(merged_outputs)
    merged_inputs = [keras_pure_input_layer_dict1, keras_pure_input_layer_dict3]

    print("inference instances: ", [{"user_id": np_user_ids[3], "item_id": np_item_ids[i]} for i in range(len(np_item_ids[:400]))])
    MovieLensWideDeepKerasModel(
        "user_book_recommend",
        keras_inputs=merged_inputs,
        keras_outputs=merged_outputs
    ).compile(
        loss="mean_squared_error",
        optimizer="adam",
        metrics=['accuracy']
    ).fit(
        loaded_dataset=[train_dataset, validation_dataset, evaluate_dataset, np_user_ids, np_item_ids, np_labels]
    ).serving(
    ).inference(tfx_host_port="http://192.168.83.111:8501/",
                model_name="user_book_recommend",
                features_list=[{"user_id": np.array([4768]).tolist(), "genres": np.array([12,28,5,7,20,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0], np.int64).tolist()} for i in range(len(np_item_ids[:400]))])
    # ).inference(tfx_host_port="http://192.168.83.111:8501/",
    #             model_name="user_book_recommend",
    #             features_list=[{"user_id": np_user_ids[3].tolist(), "item_id": np_item_ids[i].tolist()} for i in range(len(np_item_ids[:400]))])


@CommonUtils.print_exec_time
def run_hst_recommend_system2():
    """推荐系统：协同过滤"""
    from skydl.model.impl.movielens_wide_deep_keras_model import MovieLensWideDeepKerasModel
    train_dataset, validation_dataset, test_dataset, np_user_ids, np_item_ids, np_labels = MovieLensWideDeepKerasModel().load_data()
    total_users = len(np.unique(np_user_ids))
    total_items = len(np.unique(np_item_ids))
    print("******", total_users, total_items)  # 6040 3650

    # first model
    first_input = keras.layers.Input(shape=(1,), name='user_id')
    first_input_encoded = keras.layers.Embedding(input_dim=6040, output_dim=3, input_length=1)(first_input)
    first_input_encoded = keras.layers.Reshape((3,))(first_input_encoded)

    # second model
    second_input = keras.layers.Input(shape=(1,), name='item_id')
    second_input_encoded = keras.layers.Embedding(input_dim=4000, output_dim=3, input_length=1)(second_input)
    second_input_encoded = keras.layers.Reshape((3,))(second_input_encoded)

    # merged model
    merged_outputs = keras.layers.Dot(axes=1)([first_input_encoded, second_input_encoded])
    merged_outputs = keras.layers.Dense(1)(merged_outputs)

    print("inference instances: ", [{"user_id": np_user_ids[3], "item_id": np_item_ids[i]} for i in range(len(np_item_ids[:400]))])
    MovieLensWideDeepKerasModel(
        "user_book_recommend",
        keras_inputs=[first_input, second_input],
        keras_outputs=merged_outputs
    ).compile(
        loss="mean_squared_error",
        optimizer="adam",
        metrics=['accuracy']
    ).fit(
        loaded_dataset=[train_dataset, validation_dataset, test_dataset, np_user_ids, np_item_ids, np_labels]
    ).serving(
    ).inference(tfx_host_port="http://192.168.83.111:8501/",
                model_name="user_book_recommend",
                features_list=[{"user_id": np_user_ids[3], "item_id": np_item_ids[i]} for i in range(len(np_item_ids[:400]))])


@CommonUtils.print_exec_time
def keras_input_will_ok_no_need_delete():
    import pandas as pd
    df = pd.DataFrame(
        [[3, 4, 8, 0],
         [2, 6, 9, 1]],
        columns=['user_id', 'item_id', 'other_id', 'target'])

    def df_to_dataset(dataframe, shuffle=True, batch_size=32):
        dataframe = dataframe.copy()
        labels = dataframe.pop('target')
        d = dict(dataframe)
        ds = tf.data.Dataset.from_tensor_slices((d, labels))
        if shuffle:
            ds = ds.shuffle(buffer_size=len(dataframe))
        ds = ds.batch(batch_size)
        return ds
    train_ds = df_to_dataset(df)
    val_ds = df_to_dataset(df)

    # first model
    first_input = keras.layers.Input(shape=(1,), name='user_id')
    first_input_encoded = keras.layers.Embedding(input_dim=10, output_dim=3, input_length=1)(first_input)
    first_input_encoded = keras.layers.Reshape((3,))(first_input_encoded)

    # second model
    second_input = keras.layers.Input(shape=(1,), name='item_id')
    second_input_encoded = keras.layers.Embedding(input_dim=10, output_dim=3, input_length=1)(second_input)
    second_input_encoded = keras.layers.Reshape((3,))(second_input_encoded)

    merged_outputs = keras.layers.concatenate([first_input_encoded, second_input_encoded])
    merged_outputs = keras.layers.Dense(1)(merged_outputs)
    MovieLensWideDeepKerasModel(
        "user_book_xxx",
        keras_inputs=[first_input, second_input],
        keras_outputs=merged_outputs
    ).compile(
        loss="mean_squared_error",
        optimizer="adam",
        metrics=['accuracy']
    ).fit_xxx(
        loaded_dataset=[train_ds, val_ds]
    ).predict_xxx()


def nlp_ragged_tensor():
    # TensorFlow 2.0新特性之Ragged Tensor  http://www.sohu.com/a/306300588_723464
    import math
    num_buckets = 1024
    embedding_size = 16
    embedding_table = tf.Variable(tf.random.truncated_normal([num_buckets, embedding_size], stddev=1.0 / math.sqrt(embedding_size)), name="embedding_table")
    queries = tf.ragged.constant([['Who', 'is', 'Dan', 'Smith'],
                                  ['Pause'],
                                  ['Will', 'it', 'rain', 'later', 'today']
                                  ])
    word_buckets = tf.strings.to_hash_bucket_fast(queries, num_buckets)
    word_embeddings = tf.ragged.map_flat_values(tf.nn.embedding_lookup, embedding_table, word_buckets)
    marker = tf.fill([queries.nrows(), 1], '#')
    padded = tf.concat([marker, queries, marker], axis=1)  # ②
    # Build word bigrams & look up embeddings.
    bigrams = tf.strings.join([padded[:, :-1], padded[:, 1:]], separator='+')  # ③
    bigram_buckets = tf.strings.to_hash_bucket_fast(bigrams, num_buckets)
    bigram_embeddings = tf.ragged.map_flat_values(tf.nn.embedding_lookup, embedding_table, bigram_buckets)  # ④
    # Find the average embedding for each sentence
    all_embeddings = tf.concat([word_embeddings, bigram_embeddings], axis=1)  # ⑤
    avg_embedding = tf.reduce_mean(all_embeddings, axis=1)  # ⑥
    print(word_embeddings)
    print(bigram_embeddings)
    print(all_embeddings)
    print(avg_embedding)


def run_ray():
    import ray, time

    def print_something():
        print("123")

    @ray.remote
    def f():
        time.sleep(1)
        print_something()
        return 1

    ray.init()
    results = ray.get([f.remote() for i in range(4)])
    print(results)


def run_ray2():
    import ray, time
    # Define two remote functions. Invocations of these functions create tasks
    # that are executed remotely.
    @ray.remote
    def multiply(x, y):
        return np.dot(x, y)

    @ray.remote
    def zeros(size):
        return np.zeros(size)

    ray.init()

    # Start two tasks in parallel. These immediately return futures and the
    # tasks are executed in the background.
    x_id = zeros.remote((100, 100))
    y_id = zeros.remote((100, 100))
    print("x_id", ray.get(x_id))
    # Start a third task. This will not be scheduled until the first two
    # tasks have completed.
    z_id = multiply.remote(x_id, y_id)
    # Get the result. This will block until the third task completes.
    z = ray.get(z_id)
    print("z:", z)

def will_fail_not_need_delete():
    # from keras_feature_input_layer_merged_model.py
    import pandas as pd
    from tensorflow import feature_column
    from tensorflow import keras

    feature_columns = []
    child_gender_young = feature_column.categorical_column_with_vocabulary_list(
        'child_gender_young', ['boy', 'girl', ''])
    child_gender_young_one_hot = feature_column.indicator_column(child_gender_young)
    feature_columns.append(child_gender_young_one_hot)

    child_month_young = feature_column.numeric_column("child_month_young")
    kid_age_youngest_buckets = feature_column.bucketized_column(child_month_young, boundaries=[0, 12, 24, 36, 72, 96])
    feature_columns.append(kid_age_youngest_buckets)

    def keras_merged_model_fn(train_dataset, feature_columns, feature_column_names=["child_month_young", "child_gender_young"]):
        # first model
        first_input = keras.layers.Input(shape=(1,), name='first_input')
        first_input_encoded = keras.layers.Embedding(input_dim=10, output_dim=3, input_length=1)(first_input)
        first_input_encoded = keras.layers.Reshape((3,))(first_input_encoded)

        # second model
        second_input = keras.layers.Input(shape=(1,), name='second_input')
        second_input_encoded = keras.layers.Embedding(input_dim=10, output_dim=3, input_length=1)(second_input)
        second_input_encoded = keras.layers.Reshape((3,))(second_input_encoded)

        # merged model
        keras_feature_column_output_layer, keras_pure_input_layer_dict = \
            KerasFeatureInputMergedModel.build_keras_feature_column_output_layer(train_dataset.take(1),
                                                                            feature_columns,
                                                                            feature_column_names)
        merged_outputs = keras.layers.concatenate([first_input_encoded, second_input_encoded, keras_feature_column_output_layer])
        merged_outputs = keras.layers.Dense(1)(merged_outputs)

        merged_inputs = [first_input, second_input] + list(keras_pure_input_layer_dict.values())
        merged_model = keras.Model(inputs=merged_inputs, outputs=merged_outputs)
        return merged_model

    df = pd.DataFrame(
        [[3, 4, 5, 'boy', 0],
         [2, 6, 7, 'girl', 1]],
        columns=['first_input', 'second_input', 'child_month_young', 'child_gender_young', 'target'])

    def df_to_dataset(dataframe, shuffle=True, batch_size=32):
        dataframe = dataframe.copy()
        labels = dataframe.pop('target')
        d = dict(dataframe)
        ds = tf.data.Dataset.from_tensor_slices((d, labels))
        if shuffle:
            ds = ds.shuffle(buffer_size=len(dataframe))
        ds = ds.batch(batch_size)
        return ds

    train_ds = df_to_dataset(df)
    val_ds = df_to_dataset(df)

    merged_model = keras_merged_model_fn(train_ds, feature_columns, feature_column_names=["child_month_young", "child_gender_young"])
    merged_model.compile(optimizer='adam', loss='binary_crossentropy', metrics=['accuracy'])
    merged_model.fit(train_ds, validation_data=val_ds, epochs=1)


def result_is_ok_can_not_delete():
    # from: https://github.com/tensorflow/tensorflow/issues/27416
    import pandas as pd
    # !pip install tensorflow==2.0.0-alpha0
    import tensorflow as tf
    from tensorflow import feature_column
    from tensorflow import keras
    from tensorflow.keras import layers
    from sklearn.model_selection import train_test_split

    URL = 'https://storage.googleapis.com/applied-dl/heart.csv'
    dataframe = pd.read_csv(URL)
    dataframe.head()

    train, test = train_test_split(dataframe, test_size=0.2)
    train, val = train_test_split(train, test_size=0.2)
    print(len(train), 'train examples')
    print(len(val), 'validation examples')
    print(len(test), 'test examples')

    # A utility method to create a tf.data dataset from a Pandas Dataframe
    def df_to_dataset(dataframe, shuffle=True, batch_size=32):
        dataframe = dataframe.copy()
        labels = dataframe.pop('target')
        ds = tf.data.Dataset.from_tensor_slices((dict(dataframe), labels))
        if shuffle:
            ds = ds.shuffle(buffer_size=len(dataframe))
        ds = ds.batch(batch_size)
        return ds

    batch_size = 5  # A small batch sized is used for demonstration purposes
    train_ds = df_to_dataset(train, batch_size=batch_size)
    val_ds = df_to_dataset(val, shuffle=False, batch_size=batch_size)
    test_ds = df_to_dataset(test, shuffle=False, batch_size=batch_size)

    age = feature_column.numeric_column("age")

    feature_columns = []
    feature_layer_inputs = {}

    # numeric cols
    for header in ['age', 'trestbps', 'chol', 'thalach', 'oldpeak', 'slope', 'ca']:
        feature_columns.append(feature_column.numeric_column(header))
        feature_layer_inputs[header] = tf.keras.Input(shape=(1,), name=header)

    # bucketized cols
    age_buckets = feature_column.bucketized_column(age, boundaries=[18, 25, 30, 35, 40, 45, 50, 55, 60, 65])
    feature_columns.append(age_buckets)

    # indicator cols
    thal = feature_column.categorical_column_with_vocabulary_list(
        'thal', ['fixed', 'normal', 'reversible'])
    thal_one_hot = feature_column.indicator_column(thal)
    feature_columns.append(thal_one_hot)
    feature_layer_inputs['thal'] = tf.keras.Input(shape=(1,), name='thal', dtype=tf.string)

    # embedding cols
    thal_embedding = feature_column.embedding_column(thal, dimension=8)
    feature_columns.append(thal_embedding)

    # crossed cols
    crossed_feature = feature_column.crossed_column([age_buckets, thal], hash_bucket_size=1000)
    crossed_feature = feature_column.indicator_column(crossed_feature)
    feature_columns.append(crossed_feature)

    batch_size = 32
    train_ds = df_to_dataset(train, batch_size=batch_size)
    val_ds = df_to_dataset(val, shuffle=False, batch_size=batch_size)
    test_ds = df_to_dataset(test, shuffle=False, batch_size=batch_size)

    feature_layer = tf.keras.layers.DenseFeatures(feature_columns)
    feature_layer_outputs = feature_layer(feature_layer_inputs)
    x = layers.Dense(128, activation='relu')(feature_layer_outputs)
    x = layers.Dense(64, activation='relu')(x)
    baggage_pred = layers.Dense(1, activation='sigmoid')(x)

    model = keras.Model(inputs=[v for v in feature_layer_inputs.values()], outputs=baggage_pred)

    model.compile(optimizer='adam',
                  loss='binary_crossentropy',
                  metrics=['accuracy'])

    model.fit(train_ds)
    print("okkkkkk....")

def will_fail_can_delete():
    import pandas as pd
    # !pip install tensorflow==2.0.0-alpha0
    import tensorflow as tf
    from tensorflow import feature_column
    from tensorflow import keras
    from tensorflow.keras import layers
    from sklearn.model_selection import train_test_split

    URL = 'https://storage.googleapis.com/applied-dl/heart.csv'
    dataframe = pd.read_csv(URL)
    dataframe.head()

    train, test = train_test_split(dataframe, test_size=0.2)
    train, val = train_test_split(train, test_size=0.2)
    print(len(train), 'train examples')
    print(len(val), 'validation examples')
    print(len(test), 'test examples')

    # A utility method to create a tf.data dataset from a Pandas Dataframe
    def df_to_dataset(dataframe, shuffle=True, batch_size=32):
        dataframe = dataframe.copy()
        labels = dataframe.pop('target')
        ds = tf.data.Dataset.from_tensor_slices((dict(dataframe), labels))
        if shuffle:
            ds = ds.shuffle(buffer_size=len(dataframe))
        ds = ds.batch(batch_size)
        return ds

    batch_size = 5  # A small batch sized is used for demonstration purposes
    train_ds = df_to_dataset(train, batch_size=batch_size)
    val_ds = df_to_dataset(val, shuffle=False, batch_size=batch_size)
    test_ds = df_to_dataset(test, shuffle=False, batch_size=batch_size)

    age = feature_column.numeric_column("age")
    age_buckets = feature_column.bucketized_column(age, boundaries=[18, 25, 30, 35, 40, 45, 50, 55, 60, 65])
    thal = feature_column.categorical_column_with_vocabulary_list(
        'thal', ['fixed', 'normal', 'reversible'])
    thal_one_hot = feature_column.indicator_column(thal)
    thal_embedding = feature_column.embedding_column(thal, dimension=8)
    thal_hashed = feature_column.categorical_column_with_hash_bucket(
        'thal', hash_bucket_size=1000)
    crossed_feature = feature_column.crossed_column([age_buckets, thal], hash_bucket_size=1000)

    feature_columns = []

    # numeric cols
    for header in ['age', 'trestbps', 'chol', 'thalach', 'oldpeak', 'slope', 'ca']:
        feature_columns.append(feature_column.numeric_column(header))

    # bucketized cols
    age_buckets = feature_column.bucketized_column(age, boundaries=[18, 25, 30, 35, 40, 45, 50, 55, 60, 65])
    feature_columns.append(age_buckets)

    # indicator cols
    thal = feature_column.categorical_column_with_vocabulary_list(
        'thal', ['fixed', 'normal', 'reversible'])
    thal_one_hot = feature_column.indicator_column(thal)
    feature_columns.append(thal_one_hot)

    # embedding cols
    thal_embedding = feature_column.embedding_column(thal, dimension=8)
    feature_columns.append(thal_embedding)

    # crossed cols
    crossed_feature = feature_column.crossed_column([age_buckets, thal], hash_bucket_size=1000)
    crossed_feature = feature_column.indicator_column(crossed_feature)
    feature_columns.append(crossed_feature)

    batch_size = 32
    train_ds = df_to_dataset(train, batch_size=batch_size)
    val_ds = df_to_dataset(val, shuffle=False, batch_size=batch_size)
    test_ds = df_to_dataset(test, shuffle=False, batch_size=batch_size)

    feature_layer = tf.keras.layers.DenseFeatures(feature_columns)
    inputs = tf.keras.layers.Input(tensor=feature_layer, name='features')

    x = layers.Dense(128, activation='relu')(inputs)
    x = layers.Dense(64, activation='relu')(x)
    baggage_pred = layers.Dense(1, activation='sigmoid')(x)

    model = keras.Model(inputs=inputs, outputs=baggage_pred)

    model.compile(optimizer='adam',
                  loss='binary_crossentropy',
                  metrics=['accuracy'])


if __name__ == '__main__':
    # run_keras_model()
    # nlp_ragged_tensor()
    # run_movielens_wide_deep_keras_model()
    run_hst_recommend_system()
    # run_hst_recommend_system1()
    # run_hst_recommend_system2()
    run_wine_wide_deep_keras_model()
    # run_ray2()
    # result_is_ok_can_not_delete()
    # will_fail_not_need_delete() #一个例子, 如果从alpha0升级到tensorflow=2.0.0-beta1时会报错: tensorflow.python.framework.errors_impl.InvalidArgumentError:  You must feed a value for placeholder tensor 'child_gender_young'
    # keras_input_will_ok_no_need_delete()
























from setuptools import setup, find_packages
import codecs
import os
import skydl

with open("README.md") as f:
    readme = f.read()


def read(fname):
    return codecs.open(os.path.join(os.path.dirname(__file__), fname)).read()


def read_install_requires():
    with open('requirements.txt', 'r') as f:
        res = f.readlines()
    res = list(map(lambda s: s.replace('\n', ''), res))
    return res

setup(
    name='skydl',
    version=skydl.__version__,
    description="",
    long_description=readme,
    install_requires=read_install_requires(),
    setup_requires=['setuptools>=41.0.1', 'wheel>=0.33.4'],
    author='tony',
    author_email='',
    license='BSD',
    url='',
    keywords='skydl: high efficiency deep learning framework',
    classifiers=['Development Status :: 4 - Beta',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'License :: OSI Approved :: BSD License'],
    packages=find_packages(),
    package_data={'': ['*.csv']},
)


# -*- coding: utf-8 -*-
from py_common_util.common.common_utils import CommonUtils
import sys, os
from skydl.model.impl.my_pytorch_model import MyPyTorchModel
sys.path.append(os.path.abspath('%s/..' % sys.path[0]))


@CommonUtils.print_exec_time
def run_default_model():
    MyPyTorchModel(
        "my_torch_model"
    ).compile(
    ).fit()


if __name__ == '__main__':
    run_default_model()

















# -*- coding: utf-8 -*-
import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow.python import keras
import numpy as np
from skydl.datasets.movielens_dataset_builder import MovieLensDatasetBuilder
from skydl.model.default_keras_model import DefaultKerasModel
from skydl.model.train_phase_enum import TrainPhaseEnum
from py_common_util.common.common_utils import CommonUtils
import sys,os
sys.path.append(os.path.abspath('%s/..' % sys.path[0]))


class MovieLensWideDeepKerasModel2(DefaultKerasModel):
    """
    https://www.curiousily.com/posts/heart-disease-prediction-in-tensorflow-2/
    Building a book Recommendation System using Keras  https://towardsdatascience.com/building-a-book-recommendation-system-using-keras-1fba34180699
    """
    def adjust_parse_args_value(self):
        super().adjust_parse_args_value()
        self.parser_args.data_path = sys.path[0] + '/../../dataset'
        self.parser_args.use_cuda = True
        self.parser_args.init_from_saver = True
        self.parser_args.train_phase = TrainPhaseEnum.Inference.value
        self.parser_args.model_version = '1'
        self.parser_args.epochs = 1
        self.parser_args.batch_size = 128
        self.parser_args.log_interval = 1000

    @CommonUtils.print_exec_time
    def load_data(self, *args, **kwargs):
        super().load_data(*args, **kwargs)
        dataset_name = MovieLensDatasetBuilder.camelcase_to_snakecase("MovieLensDatasetBuilder")
        num_examples = MovieLensDatasetBuilder.get_num_examples(dataset_name)
        full_datasets = MovieLensDatasetBuilder.load_batched_datasets(dataset_name,
                                                                      split=[tfds.Split.TRAIN],
                                                                      batch_size=128,
                                                                      epochs=1)
        batched_train_dataset, batched_validation_dataset, batched_evaluate_dataset = MovieLensDatasetBuilder.split_full_datasets(full_datasets,
                                                                                                      num_examples=MovieLensDatasetBuilder.get_total_num_batch(
                                                                                                          num_examples,
                                                                                                          128),
                                                                                                      train_size=0.7,
                                                                                                      validation_size=0,
                                                                                                      test_size=0.2)
        np_user_ids = np.array([])
        np_item_ids = np.array([])
        np_labels = np.array([])
        if self.is_inference_phase():
            # get 2 batched test_dataset: i.e. 128 * 3
            stat_dataset = batched_evaluate_dataset.take(3)
        else:
            # get train_dataset
            stat_dataset = batched_train_dataset
        for features, labels in MovieLensDatasetBuilder.read_batched_datasets(stat_dataset):
            np_user_ids = np.append(np_user_ids, features['user_id'])
            np_item_ids = np.append(np_item_ids, features['item_id'])
            np_labels = np.append(np_labels, labels)
        return batched_train_dataset, batched_validation_dataset, batched_evaluate_dataset, np_user_ids, np_item_ids, np_labels

    def fit_ok(self, *args, **kwargs):
        """可以成功运行"""
        if not self.is_training_phase():
            return self
        train_dataset, validation_dataset, test_dataset, np_user_ids, np_item_ids, np_labels = self.load_data()
        user_feature_column = MovieLensDatasetBuilder.build_numeric_feature_columns(["user_id"])
        item_feature_column = MovieLensDatasetBuilder.build_numeric_feature_columns(["item_id"])
        total_items = len(np.unique(np_item_ids))
        total_users = len(np.unique(np_user_ids))

        # creating user embedding path
        user_input = keras.Input(shape=[1], name="User-Input")
        user_embedding = keras.layers.Embedding(total_users, 10, name="User-Embedding")(user_input)
        user_vec = keras.layers.Flatten(name="Flatten-Users")(user_embedding)

        # creating book embedding path
        book_input = keras.Input(shape=[1], name="Book-Input")
        book_embedding = keras.layers.Embedding(total_items, 10, name="Book-Embedding")(book_input)
        book_vec = keras.layers.Flatten(name="Flatten-Books")(book_embedding)

        # performing dot product and creating model
        prod = keras.layers.Dot(name="Dot-Product", axes=1)([book_vec, user_vec])
        model = keras.Model([user_input, book_input], prod)
        model.compile('adam', 'mean_squared_error', ["accuracy"])
        model.fit([np_user_ids] + [np_item_ids], np_labels)
        model.summary()
        return self

    @CommonUtils.print_exec_time
    def fit(self, *args, **kwargs):
        if not self.is_training_phase():
            return self
        loaded_dataset = kwargs.get("loaded_dataset")
        if loaded_dataset is not None:
            batched_train_dataset = loaded_dataset[0]
            batched_evaluate_dataset = loaded_dataset[2]
            self.model.get_proxy().fit(batched_train_dataset, epochs=self.parser_args.epochs, callbacks=self.fit_callbacks())
            self.model.get_proxy().summary()
            # evaluate mode
            self.evaluate(batched_evaluate_dataset)
            # save model for tf serving
            self.save_model(batched_evaluate_dataset=batched_evaluate_dataset)

        # change to inference phase
        self.parser_args.train_phase = TrainPhaseEnum.Inference.value
        return self

    @CommonUtils.print_exec_time
    def evaluate(self, *args, **kwargs):
        test_loss, test_acc = self.model.get_proxy().evaluate(args[0])
        self.model.get_proxy().summary()
        print('\nTest accuracy:', test_acc)
        return self

    @CommonUtils.print_exec_time
    def save_model(self, *args, **kwargs):
        """save model for tf serving"""
        if not self.is_training_phase():
            return self
        # restore weights from latest checkpoint
        latest_checkpoint = tf.train.latest_checkpoint(self.get_model_checkpoint_dir())
        if latest_checkpoint:
            self.model.get_proxy().load_weights(latest_checkpoint)
            self.log.info("already called load_weights from: " + latest_checkpoint)
        saved_serving_path = self.parser_args.model_checkpoint_path + "/serving/models/" + self.model.name() + "/" + self.parser_args.model_version

        batched_evaluate_dataset = kwargs.get("batched_evaluate_dataset")
        # save model for tf serving
        if not self.model.get_proxy().inputs and batched_evaluate_dataset is not None:
            for batched_examples in tfds.as_numpy(batched_evaluate_dataset.take(1)):
                test_batched_features, test_batched_labels = batched_examples
                self.model.get_proxy()._set_inputs(test_batched_features)
        tf.saved_model.save(self.model.get_proxy(), saved_serving_path)
        self.log.info("already called saved_model to saved_serving_path: " + saved_serving_path)
        return self

    @CommonUtils.print_exec_time
    def serving(self, *args, **kwargs):
        if not self.is_inference_phase():
            return self
        # restore weights from latest checkpoint
        latest_checkpoint = tf.train.latest_checkpoint(self.get_model_checkpoint_dir())
        if latest_checkpoint:
            self.model.get_proxy().load_weights(latest_checkpoint)
            self.log.info("already called load_weights from: " + latest_checkpoint)
        saved_serving_path = self.parser_args.model_checkpoint_path + "/serving/models/" + self.model.name() + "/" + self.parser_args.model_version

        _, _, evaluate_dataset, np_user_ids, np_item_ids, np_labels = self.load_data()

        # predict
        self.predict(user_id=np_user_ids[2], np_item_ids=np_item_ids)

        # save model for tf serving
        if not self.model.get_proxy().inputs:
            for batched_examples in tfds.as_numpy(evaluate_dataset.take(1)):
                test_batched_features, test_batched_labels = batched_examples
                self.model.get_proxy()._set_inputs(test_batched_features)
        tf.saved_model.save(self.model.get_proxy(), saved_serving_path)
        return self

    @CommonUtils.print_exec_time
    def predict(self, *args, **kwargs):
        """python版的预测接口，返回预测结果"""
        if not self.is_inference_phase():
            return None
        user_id = kwargs.get("user_id")
        np_item_ids = kwargs.get("np_item_ids")

        users = np.array([user_id for i in range(len(np_item_ids))])
        predictions = self.model.get_proxy().predict([users, np_item_ids])
        predictions = np.array([a[0] for a in predictions])
        recommended_book_ids = (-predictions).argsort()
        print("用户Id：", int(user_id))
        print("推荐的新闻Id列表,Top(" + str(len(np_item_ids)) + ")：", recommended_book_ids)
        print("推荐得分倒序：", predictions[recommended_book_ids])
        return None

    @CommonUtils.print_exec_time
    def inference(self, tfx_host_port="http://localhost:8501/",model_name=None, features_list=[]):
        """
        TODO 该方法将被废弃，用predict方法替代
        inference request to tf serving
        https://medium.com/tensorflow/serving-ml-quickly-with-tensorflow-serving-and-docker-7df7094aa008
        #########################################
        $docker rmi -f <<old serving images>>
        $docker pull tensorflow/serving:1.13.0
        $docker run -p 8500:8500 -p 8501:8501 \
        --mount type=bind,source=/Users/tony/myfiles/spark/share/python-projects/deep_trading/skydl/examples/saved_model/serving/models/default_keras_net,target=/models/my_keras_model \
        -e MODEL_NAME=my_keras_model -t tensorflow/serving:1.13.0
        #########################################
        :param tfx_host_port: docker serving url, e.g. "http://localhost:8501/" 完整的post url，如"http://localhost:8501/v1/models/my_keras_model:predict"
        :param model_name: inference_model.model.name() e.g. "my_keras_model"
        :param features_list:  shape e.g. (1,28,28), 数据格式：tf_dataset_batched_features.tolist()[:100]
        features_list数据可以来自如下：
        _, evaluate_dataset = inference_model.load_data()
        import tensorflow_datasets as tfds
        for dataset in tfds.as_numpy(evaluate_dataset):
            features, labels = dataset
        features_list = features.tolist()
        :return:
        """
        import json
        import numpy
        import requests
        import traceback
        data = json.dumps({"signature_name": "serving_default",
                           "instances": features_list})
        headers = {"content-type": "application/json"}
        # 去掉url里最后一个/字符
        if tfx_host_port[-1] == '/':
            tfx_host_port = tfx_host_port[:-1]
        post_url = tfx_host_port + "/v1/models/" + model_name + ":predict"
        json_response = requests.post(post_url, data=data, headers=headers)
        print("the inference request takes times(milliseconds): " + str(json_response.elapsed.total_seconds()*1000))
        try:
            if not json_response.ok:
                self.log.info("the inference request failure, url: " + json_response.url + ", response code: " + str(json_response.status_code) + ", error reason: " + json_response.text)
                return []
            prediction_list = numpy.array(json_response.json()["predictions"])
            print("response is ok, prediction_list after argmax is:" + str(tf.argmax(prediction_list, axis=1).numpy()))
            return prediction_list
        except:
            print("exception occured at an inference request!")
            traceback.print_exc()
        return []
